"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.AttributionCollection = void 0;
const common_utils_1 = require("@fluidframework/common-utils");
const collections_1 = require("./collections");
const mergeTreeNodes_1 = require("./mergeTreeNodes");
class AttributionCollection {
    constructor(baseEntry, _length) {
        this._length = _length;
        this.entries = new collections_1.RedBlackTree(mergeTreeNodes_1.compareNumbers);
        this.entries.put(0, baseEntry);
    }
    getAtOffset(offset) {
        (0, common_utils_1.assert)(offset >= 0 && offset < this._length, 0x443 /* Requested offset should be valid */);
        const node = this.entries.floor(offset);
        (0, common_utils_1.assert)(node !== undefined, 0x444 /* Collection should have at least one entry */);
        return node.data;
    }
    get length() {
        return this._length;
    }
    /**
     * Splits this attribution collection into two with entries for [0, pos) and [pos, length).
     */
    splitAt(pos) {
        const splitBaseEntry = this.getAtOffset(pos);
        const splitCollection = new AttributionCollection(splitBaseEntry, this.length - pos);
        for (let current = this.entries.ceil(pos); current !== undefined; current = this.entries.ceil(pos)) {
            // If there happened to be an attribution change at exactly pos, it's already set in the base entry
            if (current.key !== pos) {
                splitCollection.entries.put(current.key - pos, current.data);
            }
            this.entries.remove(current.key);
        }
        this._length = pos;
        return splitCollection;
    }
    append(other) {
        const lastEntry = this.getAtOffset(this.length - 1);
        other.entries.map(({ key, data }) => {
            if (key !== 0 || lastEntry !== data) {
                this.entries.put(key + this.length, data);
            }
            return true;
        });
        this._length += other.length;
    }
    getAll() {
        const results = [];
        this.entries.map(({ key, data }) => {
            results.push({ offset: key, key: data });
            return true;
        });
        return results;
    }
    clone() {
        const copy = new AttributionCollection(this.getAtOffset(0), this.length);
        this.entries.map(({ key, data }) => {
            copy.entries.put(key, data);
            return true;
        });
        return copy;
    }
    /**
     * Rehydrates attribution information from its serialized form into the provided iterable of consecutive segments.
     */
    static populateAttributionCollections(segments, summary) {
        const { keys, posBreakpoints } = summary;
        (0, common_utils_1.assert)(keys.length === posBreakpoints.length && keys.length > 0, 0x445 /* Invalid attribution summary blob provided */);
        let curIndex = 0;
        let cumulativeSegPos = 0;
        let currentInfo = keys[curIndex];
        for (const segment of segments) {
            const attribution = new AttributionCollection(currentInfo, segment.cachedLength);
            while (posBreakpoints[curIndex] < cumulativeSegPos + segment.cachedLength) {
                currentInfo = keys[curIndex];
                attribution.entries.put(posBreakpoints[curIndex] - cumulativeSegPos, currentInfo);
                curIndex++;
            }
            segment.attribution = attribution;
            cumulativeSegPos += segment.cachedLength;
        }
    }
    /**
     * Condenses attribution information on consecutive segments into a `SerializedAttributionCollection`
     */
    static serializeAttributionCollections(segments) {
        var _a;
        const posBreakpoints = [];
        const keys = [];
        let mostRecentAttributionKey;
        let cumulativePos = 0;
        let segmentsWithAttribution = 0;
        let segmentsWithoutAttribution = 0;
        for (const segment of segments) {
            if (segment.attribution) {
                segmentsWithAttribution++;
                for (const { offset, key: info } of (_a = segment.attribution.getAll()) !== null && _a !== void 0 ? _a : []) {
                    if (info !== mostRecentAttributionKey) {
                        posBreakpoints.push(offset + cumulativePos);
                        keys.push(info);
                    }
                    mostRecentAttributionKey = info;
                }
            }
            else {
                segmentsWithoutAttribution++;
            }
            cumulativePos += segment.cachedLength;
        }
        (0, common_utils_1.assert)(segmentsWithAttribution === 0 || segmentsWithoutAttribution === 0, 0x446 /* Expected either all segments or no segments to have attribution information. */);
        const blobContents = { keys, posBreakpoints, length: cumulativePos };
        return blobContents;
    }
}
exports.AttributionCollection = AttributionCollection;
//# sourceMappingURL=attributionCollection.js.map